model = """

data {

    int<lower=0> n_a;           // observations a
    int<lower=0> n_b;           // observations b
    
    vector[n_a] a;              // a vector
    vector[n_b] b;              // b vector
    
    real mu;                    // mean prior parameter
    real s;                     // standard deviation prior parameter
    real phi;                   // nu prior parameter
    
    real alpha;                 // sigma prior parameter
    real beta;                  // sigma prior parameter 
    
    }
    
parameters {

    real mu_a;                  // mu_a parameter
    real mu_b;                  // mu_b parameter 
    
    real<lower=0> sigma_a;      // sigma a parameter
    real<lower=0> sigma_b;      // sigma b parameter 
    
    real<lower=1> nu;           // nu parameter
    
    }
    
model {

    // Priors
    
    mu_a ~ normal(mu, 2 * s);
    mu_b ~ normal(mu, 2 * s);
    
    sigma_a ~ inv_gamma(alpha, beta);
    sigma_b ~ inv_gamma(alpha, beta);
    
    // exponential + 1 distribution
    target += log(phi) - ((nu - 1) * phi); 
     
    // Likelihood
    
    a ~ student_t(nu, mu_a, sigma_a);
    b ~ student_t(nu, mu_b, sigma_b);

    }

generated quantities {

    real mu_delta;              // mu_b - mu_a
    mu_delta = mu_b - mu_a;

    }

"""



import numpy as np
import arviz as az
import plotly.graph_objects as go

from bayestestimation.bayesthelpers import _calculate_kde, _calculate_map


def _get_centre_lines(draws, method):
    # Derives map or median or mean for plotting purposes
    if method == "hdi":
        cl = _calculate_map(draws)
    elif method == "quantile":
        cl = np.quantile(draws, 0.5)
    x, kde_density = _calculate_kde(draws, num=100)
    best_y = kde_density[np.argmin(np.abs(x - cl))]
    return {"x": cl, "y": best_y}


def _get_intervals(draws, method, bounds):
    # Derives HDI or credible intervals for plotting purposes
    if method == "hdi":
        il = az.hdi(draws, bounds)
    elif method == "quantile":
        il = np.quantile(draws, bounds)
    x, kde_density = _calculate_kde(draws, num=100)
    subx = x[(x > il[0]) & (x < il[1])]
    kde_density = kde_density[(x > il[0]) & (x < il[1])]
    return {"x": subx, "y": kde_density}


def _make_density_go(draws, name, col="#000000"):
    # Makes a KDE density graph object
    x, kde_density = _calculate_kde(draws, num=100)
    graphobj = go.Scatter(x=x, y=kde_density, line={"color": col}, name=name)
    return graphobj


def _make_histogram_go(draws, name, col="#000000"):
    # Makes a histogram graph object
    graphobj = go.Histogram(
        x=draws,
        nbinsx=40,
        opacity=0.2,
        histnorm="probability density",
        marker_color=col,
        name=name,
    )
    return graphobj


def _make_area_go(kde_object, name, col="#000000"):
    # Makes area under curve graph object
    graphobj = go.Scatter(
        x=kde_object["x"],
        y=kde_object["y"],
        fill="tozeroy",
        line={"color": col},
        opacity=0.5,
        name=name,
    )
    return graphobj


def _make_line_go(line_object, name, col="#000000"):
    # Make line graph object
    graphobj = go.Scatter(
        x=[line_object["x"], line_object["x"]],
        y=[0, line_object["y"]],
        mode="lines",
        line={"color": col, "dash": "dash"},
        name=name,
    )
    return graphobj


def _make_delta_line(draws, delta_line, col="#d62728"):
    # Make line dictionary object for the delta posterior
    x, kde_density = _calculate_kde(draws, num=100)
    maxy = np.max(kde_density) * 1.2
    linedict = {
        "type": "line",
        "x0": delta_line,
        "x1": delta_line,
        "y0": 0,
        "y1": maxy,
        "xref": "x3",
        "yref": "y3",
        "line": dict(color=col, dash="dot"),
    }
    return linedict



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "numpy>=1.17.2",
    "scipy>=1.3.1",
    "pandas>=0.25.1",
    "arviz>=0.9.0",
    "pystan>=2.19.1.1",
    "plotly>=4.9.0",
]

setup_requirements = [
    "pytest-runner",
    "numpy>=1.17.2",
    "scipy>=1.3.1",
    "pandas>=0.25.1",
    "arviz>=0.9.0",
    "pystan>=2.19.1.1",
    "plotly>=4.9.0",
]

test_requirements = [
    "pytest>=3",
    "numpy>=1.17.2",
    "scipy>=1.3.1",
    "pandas>=0.25.1",
    "arviz>=0.9.0",
    "pystan>=2.19.1.1",
    "plotly>=4.9.0",
]

long_description = """

Class method for Bayesian estimation and comparison of means

* Free software: MIT license

Features
--------

Class method that acts as a wrapper for a pystan formulation of a Bayesian 't-test' using the BEST implementation.

- Estimates the posterior distributions of the mean and standard deviation parameters for two samples, A and B

- Estimates of the posterior distribution of the difference in mean parameters for two samples, A and B.

- Provides summary statistics and visualisations for the estimated parameters.

- The prior parameters, sample count, random seed, credible intervals, HDI and parameter names can all be customised.

- The stan model object and stan model fit object can be accessed just like using pystan directly
"""

setup(
    author="Oliver Chipperfield",
    author_email="omc0dev@gmail.com",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Class method for Bayesian estimation and comparison of means",
    install_requires=requirements,
    license="MIT license",
    long_description=long_description,
    include_package_data=True,
    keywords="bayestestimation",
    name="bayestestimation",
    packages=find_packages(include=["bayestestimation", "bayestestimation.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/oli-chipperfield/bayestestimation",
    version="0.9.0",
    zip_safe=False,
)



import numpy as np
import pandas as pd
import arviz as az
import pystan as pystan

from plotly.subplots import make_subplots
from bayestestimation.best import model
from bayestestimation.bayesthelpers import _calculate_map
from bayestestimation.bayestplotters import _get_centre_lines
from bayestestimation.bayestplotters import _get_intervals
from bayestestimation.bayestplotters import _make_density_go
from bayestestimation.bayestplotters import _make_histogram_go
from bayestestimation.bayestplotters import _make_area_go
from bayestestimation.bayestplotters import _make_line_go
from bayestestimation.bayestplotters import _make_delta_line


class BayesTEstimation:
    def __init__(self):
        """
        Initialises the BayesTEstimation class and compiles the stan model
        """
        self.stan_model = self._compile_model()
        par_list = ["mu_a", "mu_b", "mu_delta", "sigma_a", "sigma_b", "nu"]
        self.par_list = par_list

    def _compile_model(self):
        # Compiles the stan model in C++
        return pystan.StanModel(model_code=model, model_name="BEST")

    def _check_sample_posterior_inputs(self):
        # Checks that parameters are in the correct format
        types = ["list", "ndarray", "Series"]
        mus_types = ["float", "int"]
        if (type(self.a).__name__ not in types) or (type(self.b).__name__ not in types):
            raise ValueError(
                "type(a).__name__ and/or type(b).__name__ must be 'list', 'ndarray' or 'Series'"
            )
        if self.n is not None:
            if self.n <= 0 or str(self.n).isdigit() == False:
                raise ValueError("n must be a positive integer")
        if (self.prior_alpha <= 0) or (self.prior_beta <= 0) or (self.prior_phi <= 0):
            raise ValueError(
                "the prior_alpha and/or prior_beta and/or prior_phi parameters must be > 0"
            )
        if self.prior_mu is not None and (
            type(self.prior_mu).__name__ not in mus_types
        ):
            raise ValueError(
                "prior_mu must be None or type(prior_mu).__name__ must be 'float' or 'int'"
            )
        if self.prior_s is not None:
            if type(self.prior_s).__name__ not in mus_types:
                raise ValueError("prior_s must be None or must be > 0")
            elif self.prior_s <= 0:
                raise ValueError("prior_s must be None or must be > 0")
        if self.seed is not None:
            if str(self.seed).isdigit() == False:
                raise ValueError("seed must be a positive integer or None")

    def _estimate_prior_mu(self):
        # Estimates a prior value for mu when one isn't provided
        return np.mean(np.concatenate([self.a, self.b]))

    def _estimate_prior_s(self):
        # Estimates a prior value for s when one isn't provided
        return np.std(np.concatenate([self.a, self.b]))

    def _build_input_dictionary(self):
        # Builds a dictionary input to apply to the model
        return {
            "a": self.a,
            "b": self.b,
            "n_a": self.n_a,
            "n_b": self.n_b,
            "mu": self.prior_mu,
            "s": self.prior_s,
            "phi": self.prior_phi,
            "alpha": self.prior_alpha,
            "beta": self.prior_beta,
        }

    def _extract_posteriors(self):
        # Extracts samples of the posteriors from the stan model
        self.unpermuted_extract = {}
        for i in self.par_list:
            self.unpermuted_extract[i] = self.fit.extract(pars=i, permuted=False)[i]

    def _flatten_extracts(self):
        # Flattens the chains of the posteriors
        dic = {}
        for i in self.par_list:
            dic[i] = self.unpermuted_extract[i].flatten()
        return dic

    def fit_posteriors(
        self,
        a,
        b,
        n=10000,
        prior_alpha=0.001,
        prior_beta=0.001,
        prior_phi=(1 / 30),
        prior_mu=None,
        prior_s=None,
        seed=None,
    ):
        """
        Fits the data and prior parameters and samples from the posterior distribution.
        Parameters
        ----------
        a: list, ndarray or Series, array of continuous data containing the results from sample a.
        b: list, ndarray or Series, array of continuous data containing the results from sample b.
        n: int, the total posterior number of posterior samples to take (after burn-in).  Default 10000.
        prior_alpha: float > 0, the alpha parameter for the inv-gamma prior distribution of the standard deviation of a and b.  Default = 0.001.
        prior_beta: float > 0, the beta parameter for the inv-gamma prior distribution of the standard deviation of a and b.  Default = 0.001.
        prior_phi: float > 0, the phi parameter for the nu prior.  Default = (1/30).
        prior_mu: float, the mean parameter for the mu priors for a and b.  Default is None.  If None the combined sample mean is used.
        prior_s: float, the standard deviation parameter for the mu priors for a and b.  Default is None.  If None the combined standard deviation is used.
        seed: integer > 0, set random seed at the start of sampling, default = None
        """
        self.a = a
        self.b = b
        self.n_a = len(a)
        self.n_b = len(b)
        self.n = n
        self.prior_alpha = prior_alpha
        self.prior_beta = prior_beta
        self.prior_phi = prior_phi
        self.prior_mu = prior_mu
        self.prior_s = prior_s
        self.seed = seed
        self._check_sample_posterior_inputs()
        if self.prior_mu is None:
            self.prior_mu = self._estimate_prior_mu()
        if self.prior_s is None:
            self.prior_s = self._estimate_prior_s()
        self.fit = self.stan_model.sampling(
            data=self._build_input_dictionary(),
            seed=self.seed,
            iter=int(np.ceil((self.n / 4) * 2)),
        )
        self._extract_posteriors()

    def _check_for_fit(self):
        # Checks to see if the self.fit exists
        if "fit" not in list(self.__dict__.keys()):
            raise NameError(
                "You must add data using the fit_posteriors method before using this method"
            )

    def get_posteriors(self):
        """
        Retrieves random draws from the posteriors
        Returns
        -------
        dictionary:
            - mu_a: draws from the posterior of mu_a.
            - mu_b: draws from the posterior of mu_b.
            - mu_delta: draws from the posterior of mu_delta.
            - sigma_a: draws from the posterior of sigma_a.
            - sigma_b: draws from the posterior of sigma_b.
            - nu: draws from the posterior of nu.
        """
        self._check_for_fit()
        return self._flatten_extracts()

    def _calculate_quantiles(self, d, mean, quantiles):
        # Calculate mean and quantiles
        q = np.quantile(d, quantiles)
        if mean is True:
            q = np.append(q, np.mean(d))
        return q

    def quantile_summary(self, mean=True, quantiles=[0.025, 0.5, 0.975], names=None):
        """
        Summarises the properties of the estimated posterior using quantiles
        Parameters
        ----------
        mean:  boolean, default True, calculates the mean of the draws from the posterior.  Default True
        quantiles: list, calculates the quantiles of the draws from the posterior.  Default [0.025, 0.5, 0.975]
        names:  list of length 6, parameter names in order: a, b, b-a.  Default ['mu_a', 'mu_b', 'mu_delta', 'sigma_a', 'sigma_b', 'nu']
        Returns
        -------
        pd.DataFrame:
            'mu_a':  summaries of the posterior of mu_a
            'mu_b':  summaries of the posterior of mu_b
            'mu_delta':  summaries of the posterior of mu_b - mu_a
            'sigma_a': Summaries of the posterior of sigma_a
            'sigma_b':  Summaries of the posterior of sigma_b
            'nu':  Summaries of the posterior of nu.

        """
        self._check_for_fit()
        if quantiles is None:
            raise ValueError("quantiles must be a list of length > 0")
        draws = list(self._flatten_extracts().values())
        if names is None:
            names = self.par_list
        if len(names) != 6:
            raise ValueError("names must be a list of length 6")
        q = []
        for i in draws:
            q.append(self._calculate_quantiles(i, mean, quantiles))
        df = pd.DataFrame(np.array(q))
        if mean is True:
            df.columns = list(map(str, quantiles)) + ["mean"]
        else:
            df.columns = list(map(str, quantiles))
        df["parameter"] = names
        return df

    def _calculate_hdi_and_map(self, d, mean, interval):
        # Calculate HDI interval and MAP
        q = az.hdi(d, hdi_prob=interval)
        m = _calculate_map(d)
        q = np.array([q[0], m, q[1]])
        if mean is True:
            q = np.append(q, np.mean(d))
        return q

    def hdi_summary(self, mean=True, interval=0.95, names=None):
        """
        Summarises the properties of the estimated posteriors using the MAP and HDI
        Parameters
        ----------
        mean:  boolean, calculates the mean of the draws from the posterior.  Default True
        interval: float, defines the HDI interval.  Default = 0.95 (i.e. 95% HDI interval)
        names:  list of length 6, parameter names in order: a, b, b-a.  Default ['mu_a', 'mu_b', 'mu_delta', 'sigma_a', 'sigma_b', 'nu']
        Returns
        -------
        pd.DataFrame:
            'mu_a': Summaries of the posterior of mu_a
            'mu_b': Summaries of the posterior of mu_b
            'mu_delta': Summaries of the posterior of mu_b - mu_a
            'sigma_a': Summaries of the posterior of sigma_a
            'sigma_b': Summaries of the posterior of sigma_b
            'nu':  Summaries of the posterior of nu.
        """
        self._check_for_fit()
        if interval is None or interval <= 0 or interval >= 1:
            raise ValueError("interval must be a float > 0 and < 1")
        draws = list(self._flatten_extracts().values())
        if names is None:
            names = self.par_list
        if len(names) != 6:
            raise ValueError("names must be a list of length 6")
        q = []
        for i in draws:
            q.append(self._calculate_hdi_and_map(i, mean, interval))
        df = pd.DataFrame(np.array(q))
        col_names = [
            "%.5g" % ((1 - interval) / 2),
            "MAP",
            "%.5g" % (interval + ((1 - interval) / 2)),
        ]
        if mean is True:
            df.columns = col_names + ["mean"]
        else:
            df.columns = col_names
        df["parameter"] = names
        return df

    def _probability_interpretation_guide(self, p):
        # Interpretation guide for probabilities using:
        # https://www.cia.gov/library/center-for-the-study-of-intelligence/csi-publications/books-and-monographs/sherman-kent-and-the-board-of-national-estimates-collected-essays/6words.html
        if p >= 0 and p <= 0.13:
            i = "almost certainly not"
        elif p > 0.13 and p <= 0.4:
            i = "probably not"
        elif p > 0.4 and p <= 0.6:
            i = "about equally likely"
        elif p > 0.6 and p <= 0.86:
            i = "probably"
        elif p > 0.86 and p <= 1:
            i = "almost certainly"
        else:
            raise ValueError("p must be >= 0 and <= 1")
        return i

    def _print_inference_probability(self, p, i, direction, value, names):
        # Combines inference values into a readable string
        s = "The probability that " + names[1] + " is " + direction + " " + names[0]
        if value != 0:
            s = s + " by more than " + str(value)
        s = s + " is " + ("%.5g" % (p * 100)) + "%."
        s = s + " Therefore " + names[1] + " is " + i + " " + direction + " " + names[0]
        if value != 0:
            s = s + " by more than " + str(value)
        s = s + "."
        return s

    def infer_delta_probability(
        self, direction="greater than", value=0, print_inference=True, names=None
    ):
        """
        Provides a guide to making inferences on the posterior delta, based on proportion of
        draws to the right or left of a given value.
        Parameters
        ----------
        direction: str, defines the direction of the inference, options 'greater than' or 'less than'.  Default is 'greater than'.
        value: float,  defines the value about which to make the inference.  Default = 0.
        print_inference:  boolean, prints a readable string.  Default is True.
        names:  list of length 3, parameter names in order: a, b, b-a.  Default ['mu_a', 'mu_b', 'mu_delta']
        Returns
        -------
        tuple
            - float, probability that b > (a + value) or b < (a + value).
            - str, string interpretation of that probability
        """
        self._check_for_fit()
        dir_opts = ["greater than", "less than"]
        if direction not in dir_opts:
            raise ValueError("direction must be 'greater than' or 'less than'")
        d = self.unpermuted_extract["mu_delta"].flatten()
        if direction == "greater than":
            p = len(d[d > 0]) / len(d)
        else:
            p = len(d[d < 0]) / len(d)
        i = self._probability_interpretation_guide(p)
        if names is None:
            names = ["mu_a", "mu_b", "mu_delta"]
        if len(names) != 3:
            raise ValueError("names must be a list of length 3")
        if print_inference is True:
            print(self._print_inference_probability(p, i, direction, value, names))
        return p, i

    def _bayes_factor_interpretation_guide(self, bf):
        # Interpretation guide for bayes factors using:
        # Jeffreys guide (https://en.wikipedia.org/wiki/Bayes_factor#cite_note-9)
        if np.isinf(bf) or bf > np.power(10, 2):
            i = "decisive"
        elif bf > np.power(10, 3 / 2) and bf <= np.power(10, 2):
            i = "very strong"
        elif bf > 10 and bf <= np.power(10, 3 / 2):
            i = "strong"
        elif bf > np.power(10, 1 / 2) and bf <= 10:
            i = "substantial"
        elif bf >= 1 and bf <= np.power(10, 1 / 2):
            i = "barely worth mentioning"
        elif bf < 1:
            i = "negative"
        else:
            raise ValueError("bf did not satisfy range of criteria")
        return i

    def _print_inference_bayes_factor(self, bf, i, direction, value, names):
        s = (
            "The calculated bayes factor for the hypothesis that "
            + names[1]
            + " is "
            + direction
            + " "
            + names[0]
        )
        if value != 0:
            s = s + " by more than " + str(value)
        s = (
            s
            + " versus the hypothesis that "
            + names[0]
            + " is "
            + direction
            + " "
            + names[0]
        )
        if value != 0:
            s = s + " by more than " + str(value)
        s = s + " is "
        if np.isinf(bf) is True:
            s = s + "more than 100"
        else:
            s = s + ("%.5g" % bf)
        s = s + ". Therefore the strength of evidence for this hypothesis is " + i + "."
        return s

    def _estimate_bayes_factor(self, p_h1, p_h2):
        # Estimates bayes Factor
        if p_h2 == 0:
            k = np.Infinity
        else:
            k = p_h1 / p_h2
        return k

    def infer_delta_bayes_factor(
        self, direction="greater than", value=0, print_inference=True, names=None
    ):
        """
        Provides a guide to making inferences on the posterior delta, based on the Bayes Factor by estimating
        P(D|H1) / P(D|H2) for the hypotheses H1: b>(a + value) vs H2: (a + value)>b (or vice versa).
        Where D denotes the observed data.
        Parameters
        ----------
        direction: str, defines the direction of the inference, options 'greater than' or 'less than'.  Default is 'greater than'.
        value: float,  defines the value about which to make the inference.  Default = 0.
        print_inference:  boolean, prints a readable string.  Default is True.
        names:  list of length 3, parameter names in order: a, b, b-a.  Default ['mu_a', 'mu_b', 'mu_delta']
        Returns
        -------
        tuple
            - float, bayes factor for P(D|H1) / P(D|H2) for the hypotheses H1: b>(a + value) vs H2: (a + value)>b (or vice versa).
            - str, string interpretation of that bayes factor
        """
        self._check_for_fit()
        dir_opts = ["greater than", "less than"]
        if direction not in dir_opts:
            raise ValueError("direction must be 'greater than' or 'less than'")
        d = self.unpermuted_extract["mu_delta"].flatten()
        if direction == "greater than":
            p_h1 = len(d[d > value]) / len(d)
            p_h2 = 1 - p_h1
            bf = self._estimate_bayes_factor(p_h1, p_h2)
        else:
            p_h1 = len(d[d < value]) / len(d)
            p_h2 = 1 - p_h1
            bf = self._estimate_bayes_factor(p_h1, p_h2)
        i = self._bayes_factor_interpretation_guide(bf)
        if names is None:
            names = ["mu_a", "mu_b", "mu_delta"]
        if len(names) > 3:
            raise ValueError("names must be a list of length 3")
        if print_inference is True:
            print(self._print_inference_bayes_factor(bf, i, direction, value, names))
        return bf, i

    def posterior_plot(
        self,
        method="hdi",
        delta_line=0,
        col="#1f77b4",
        bounds=None,
        names=None,
        fig_size=None,
    ):
        """
        Plots the density of the draws from the posterior distribution
        Parameters
        ----------
        method: str, defines method for interval estimate and central tendency.  Default = 'hdi'
            - 'hdi':  Uses HDI and maximum aposteriori
            - 'quantile': Uses credible intervals and median
        delta_line: float, position of the vertical line on the delta plot
        col:  str, colour of plots.  Default = '#1f77b4' (muted-blue)
        bounds:  float or list, defines the boundaries of the interval
            - if method = 'hdi': float, defines the interval of the HDI. Default = 0.95
            - if method = 'quantile': list, defines the credible interval.  Default = [0.025, 0.975]
        names: list of length 3, parameter names for the plot.  Default ['theta_a', 'theta_b', 'delta']
        fig_size:  tuple(width, height), dimensions of plot.  Default is None
        """
        self._check_for_fit()
        valid_methods = ["hdi", "quantile"]
        if method not in valid_methods:
            raise ValueError("method must be 'hdi' or 'quantile'")
        if method == "hdi" and bounds is None:
            bounds = 0.95
        if method == "quantile" and bounds is None:
            bounds = [0.025, 0.975]
        if method == "hdi" and (bounds <= 0 or bounds >= 1):
            raise ValueError(
                "if method is 'hdi' then bounds must be a float between 0 and 1"
            )
        if method == "quantiles" and len(bounds) != 2:
            raise ValueError("quantiles must be a list of length 2")
        if names is None:
            names = self.par_list
        if len(names) != 6:
            raise ValueError("names must be a list of length 6")
        if method == "hdi":
            interval_name = "hdi"
            centre_line_name = "map"
        else:
            interval_name = "credible interval"
            centre_line_name = "median"
        fig = make_subplots(
            rows=2,
            cols=3,
            shared_xaxes=False,
            shared_yaxes=False,
            subplot_titles=tuple(names),
        )
        draws = list(self._flatten_extracts().values())
        sp_coords = np.column_stack([np.tile([1, 2, 3], 2), np.repeat([1, 2], 3)])
        for i in range(0, 6):
            x = sp_coords[i][0]
            y = sp_coords[i][1]
            cl = _get_centre_lines(draws[i], method=method)
            intervals = _get_intervals(draws[i], method=method, bounds=bounds)
            fig.add_trace(
                _make_density_go(draws[i], name="posterior density", col=col), y, x
            )
            fig.add_trace(
                _make_histogram_go(draws[i], name="posterior draws", col=col), y, x
            )
            fig.add_trace(_make_line_go(cl, name=centre_line_name, col=col), y, x)
            fig.add_trace(_make_area_go(intervals, name=interval_name, col=col), y, x)
        fig.update_layout(
            shapes=[
                _make_delta_line(
                    self._flatten_extracts()["mu_delta"], delta_line=delta_line
                )
            ]
        )
        fig.update_yaxes(title_text="density", row=1, col=1)
        name_set = set()
        fig.for_each_trace(
            lambda trace: trace.update(showlegend=False)
            if (trace.name in name_set)
            else name_set.add(trace.name)
        )
        if fig_size is not None:
            fig.update_layout(height=fig_size[1], width=fig_size[0])
        return fig

    def get_rhat(self):
        """
        Extracts and summarises the rhat convergence statistics for each parameter.
        Returns
        -------
        pd.DataFrame:
            'parameters: parameter names as used in model fitting.
            'rhat': rhat statistics for each parameter.
        """
        self._check_for_fit()
        fit_summary = self.fit.summary()
        rhat_index = fit_summary["summary_colnames"].index("Rhat")
        return pd.DataFrame(
            {
                "parameters": fit_summary["summary_rownames"],
                "rhat": fit_summary["summary"][:, rhat_index],
            }
        )



#!/usr/bin/env python
import pytest
import numpy as np
import pandas as pd
from bayestestimation.bayestestimation import BayesTEstimation
from bayestestimation.bayesthelpers import _calculate_kde
from bayestestimation.bayesthelpers import _calculate_map
from bayestestimation.bayestplotters import _get_centre_lines
from bayestestimation.bayestplotters import _get_intervals
from bayestestimation.bayestplotters import _make_density_go
from bayestestimation.bayestplotters import _make_histogram_go
from bayestestimation.bayestplotters import _make_area_go
from bayestestimation.bayestplotters import _make_line_go
from bayestestimation.bayestplotters import _make_delta_line


def make_random_numpy(seed, size):
    np.random.seed(seed)
    return np.random.normal(size=size)


# Define fit_posteriors fixtures


@pytest.fixture
def make_a_numpy():
    return make_random_numpy(2020, 20)


@pytest.fixture
def make_a_list():
    return list(make_random_numpy(2020, 20))


@pytest.fixture
def make_a_series():
    return pd.DataFrame({"a": make_random_numpy(2020, 20)})["a"]


@pytest.fixture
def make_b_numpy():
    return make_random_numpy(2021, 21)


@pytest.fixture
def make_b_list():
    return list(make_random_numpy(2021, 21))


@pytest.fixture
def make_b_series():
    return pd.DataFrame({"a": make_random_numpy(2021, 21)})["a"]


@pytest.fixture
def make_explicit_n():
    return 3


@pytest.fixture
def make_explicit_prior():
    return 1


@pytest.fixture
def make_explicit_seed():
    return 1000


# Define fit_posteriors bad fixtures


@pytest.fixture
def make_a_str():
    return "foo"


@pytest.fixture
def make_b_str():
    return "bar"


@pytest.fixture
def make_invalid_n():
    return -1


@pytest.fixture
def make_invalid_prior_alpha_beta():
    return -1


@pytest.fixture
def make_invalid_prior_phi():
    return -1


@pytest.fixture
def make_invalid_prior_mu():
    return "foobar"


@pytest.fixture
def make_invalid_prior_s_1():
    return "bar"


@pytest.fixture
def make_invalid_prior_s_2():
    return -1


@pytest.fixture
def make_invalid_seed():
    return 1.2


# Define get_posteriors fixtures
@pytest.fixture
def make_get_posteriors_results():
    return np.array(
        [
            [1.82178765, 0.26726264, 0.51371462, -0.26307566],
            [0.22842242, -0.75733768, 1.51778432, 0.28337746],
            [-1.59336522, -1.02460032, 1.0040697, 0.54645312],
            [1.31736902, 1.19313651, 4.29328488, 0.26494139],
            [0.32062644, 0.40769547, 4.28318156, 1.30801229],
            [6.12807855, 3.17060426, 2.22557827, 1.55142898],
        ]
    )


@pytest.fixture
def make_get_posteriors_results_with_explicit_prior_alpha():
    return np.array(
        [
            [1.82178765, 0.26726264, 0.51371462, -0.26307566],
            [0.22842242, -0.75733768, 1.51778432, 0.28337746],
            [-1.59336522, -1.02460032, 1.0040697, 0.54645312],
            [1.30712739, 1.0530709, 4.16132443, 0.25679802],
            [0.31813379, 0.35983497, 4.15153165, 1.2678086],
            [6.12807855, 3.17060426, 2.22557827, 1.55142898],
        ]
    )


@pytest.fixture
def make_get_posteriors_results_with_explicit_prior_beta():
    return np.array(
        [
            [1.82178765, 0.26726264, 0.51371462, -0.26307566],
            [0.22842242, -0.75733768, 1.51778432, 0.28337746],
            [-1.59336522, -1.02460032, 1.0040697, 0.54645312],
            [1.328383, 1.28577873, 4.72220512, 0.30623596],
            [0.33403218, 0.4156223, 4.30486861, 1.33270051],
            [6.12807855, 3.17060426, 2.22557827, 1.55142898],
        ]
    )


@pytest.fixture
def make_get_posteriors_results_with_explicit_prior_phi():
    return np.array(
        [
            [1.82178765, 0.26726264, 0.51371462, -0.26307566],
            [0.22842242, -0.75733768, 1.51778432, 0.28337746],
            [-1.59336522, -1.02460032, 1.0040697, 0.54645312],
            [1.31736902, 1.19313651, 4.29328488, 0.26494139],
            [0.32062644, 0.40769547, 4.28318156, 1.30801229],
            [5.90391579, 2.76084894, 2.09393342, 1.53988395],
        ]
    )


@pytest.fixture
def make_get_posteriors_results_with_explicit_prior_mu():
    return np.array(
        [
            [1.82490197, 0.31709179, 0.52617191, -0.25061837],
            [0.23153675, -0.70750852, 1.53024161, 0.29583474],
            [-1.59336522, -1.02460032, 1.0040697, 0.54645312],
            [1.31736902, 1.19313651, 4.29328488, 0.26494139],
            [0.32062644, 0.40769547, 4.28318156, 1.30801229],
            [6.12807855, 3.17060426, 2.22557827, 1.55142898],
        ]
    )


@pytest.fixture
def make_get_posteriors_results_with_explicit_prior_s():
    return np.array(
        [
            [1.82318499, 0.28187739, 0.51920123, -0.26547989],
            [0.22860125, -0.75610507, 1.52251148, 0.2834922],
            [-1.59458373, -1.03798245, 1.00331025, 0.54897209],
            [1.31736902, 1.19313651, 4.29328488, 0.26494139],
            [0.32062644, 0.40769547, 4.28318156, 1.30801229],
            [6.12807855, 3.17060426, 2.22557827, 1.55142898],
        ]
    )


# Fixtures for testing results summaries


@pytest.fixture
def make_quantile_summary():
    return np.array(
        [
            [-0.7514505353555653, -0.3606663387131223, 0.039108253763252415],
            [-0.4744652899078007, -0.05335698459311282, 0.35343529956443226],
            [-0.26205676245663007, 0.3079240594731508, 0.8664490917659036],
            [0.6091506486807179, 0.8405139282720744, 1.2092274039739142],
            [0.6503299704904818, 0.8887172883934797, 1.2715059996905795],
            [5.370919434492186, 29.918093516704694, 124.19139809164547],
        ]
    )


@pytest.fixture
def make_hdi_summary():
    return np.array(
        [
            [-0.757382144261888, -0.34234680836424614, 0.029564917758929526],
            [-0.4555505896647165, -0.035785538079082135, 0.3699776818490415],
            [-0.24967564966754063, 0.3219291023903855, 0.8778163616507515],
            [0.5778856275916361, 0.8209193793911209, 1.1627010288601225],
            [0.6247454942357719, 0.8577478205548237, 1.228128295339814],
            [2.4371793019105104, 15.978161558029015, 103.45486599210501],
        ]
    )


@pytest.fixture
def make_get_rhat():
    return np.array(
        [
            0.9999232374392668,
            1.000098518503359,
            1.000104301007721,
            0.9997725213816185,
            0.9998265674210559,
            1.0001492775561391,
            1.0006674973934064,
        ]
    )


# Define fixtures for delta inference testing


@pytest.fixture
def make_infer_delta_probability_result():
    return (0.857, "probably")


@pytest.fixture
def make_infer_delta_bayes_factor_result():
    return (5.9930069930069925, "substantial")


# Define fixtures for bayeprophelpers testing


@pytest.fixture
def make_draw():
    np.random.seed(1000)
    return np.random.beta(2, 5, 100)


@pytest.fixture
def _calculate_kde_results():
    return (
        np.array([0.03857532, 0.43082818, 0.82308105]),
        np.array([1.14415602, 1.12407264, 0.11679198]),
    )


@pytest.fixture
def _calculate_map_results():
    return 0.038575315797957296


# Run helper tests


def test__calculate_kde_returns_correct_values(make_draw, _calculate_kde_results):
    x, kde_density = _calculate_kde(make_draw, num=3)
    assert np.allclose(x, _calculate_kde_results[0])
    assert np.allclose(kde_density, _calculate_kde_results[1])


def test__calculate_map_results(make_draw, _calculate_map_results):
    assert np.isclose(_calculate_map_results, _calculate_map(make_draw, num=3))


# Run initialisation test


def test_BayesTEstimation_initialises():
    try:
        BayesTEstimation()
    except:
        raise pytest.fail()


# Run fit_posteriors handles bad inputs


def test_BayesTEstimattion_handles_initialises_and_inputs_handled_correctly(
    make_a_numpy,
    make_b_numpy,
    make_a_str,
    make_b_str,
    make_invalid_n,
    make_invalid_prior_alpha_beta,
    make_invalid_prior_phi,
    make_invalid_prior_mu,
    make_invalid_prior_s_1,
    make_invalid_prior_s_2,
    make_invalid_seed,
):
    # Initialisation
    BayesTEstimationClass = BayesTEstimation()
    # Test good inputs
    try:
        BayesTEstimationClass.fit_posteriors(a=make_a_numpy, b=make_b_numpy)
    except:
        raise pytest.fail()
    # Test bad inputs
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(a=make_a_str, b=make_b_numpy)
    assert (
        str(e.value)
        == "type(a).__name__ and/or type(b).__name__ must be 'list', 'ndarray' or 'Series'"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(a=make_a_numpy, b=make_b_str)
    assert (
        str(e.value)
        == "type(a).__name__ and/or type(b).__name__ must be 'list', 'ndarray' or 'Series'"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, n=make_invalid_n
        )
    assert str(e.value) == "n must be a positive integer"
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_alpha=make_invalid_prior_alpha_beta
        )
    assert (
        str(e.value)
        == "the prior_alpha and/or prior_beta and/or prior_phi parameters must be > 0"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_beta=make_invalid_prior_alpha_beta
        )
    assert (
        str(e.value)
        == "the prior_alpha and/or prior_beta and/or prior_phi parameters must be > 0"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_phi=make_invalid_prior_phi
        )
    assert (
        str(e.value)
        == "the prior_alpha and/or prior_beta and/or prior_phi parameters must be > 0"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_mu=make_invalid_prior_mu
        )
    assert (
        str(e.value)
        == "prior_mu must be None or type(prior_mu).__name__ must be 'float' or 'int'"
    )
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_s=make_invalid_prior_s_1
        )
    assert str(e.value) == "prior_s must be None or must be > 0"
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, prior_s=make_invalid_prior_s_2
        )
    assert str(e.value) == "prior_s must be None or must be > 0"
    with pytest.raises(ValueError) as e:
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, seed=make_invalid_seed
        )
    assert str(e.value) == "seed must be a positive integer or None"


def test_BayesTEstimattion_outputs_correct_results(
    make_a_numpy,
    make_b_numpy,
    make_explicit_n,
    make_explicit_prior,
    make_explicit_seed,
    make_get_posteriors_results,
    make_get_posteriors_results_with_explicit_prior_alpha,
    make_get_posteriors_results_with_explicit_prior_beta,
    make_get_posteriors_results_with_explicit_prior_phi,
    make_get_posteriors_results_with_explicit_prior_mu,
    make_get_posteriors_results_with_explicit_prior_s,
):
    # Initialisation
    BayesTEstimationClass = BayesTEstimation()
    # Test outputs
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy, b=make_b_numpy, n=make_explicit_n, seed=make_explicit_seed
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results)
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy,
        b=make_b_numpy,
        prior_alpha=make_explicit_prior,
        n=make_explicit_n,
        seed=make_explicit_seed,
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results_with_explicit_prior_alpha)
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy,
        b=make_b_numpy,
        prior_beta=make_explicit_prior,
        n=make_explicit_n,
        seed=make_explicit_seed,
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results_with_explicit_prior_beta)
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy,
        b=make_b_numpy,
        prior_phi=make_explicit_prior,
        n=make_explicit_n,
        seed=make_explicit_seed,
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results_with_explicit_prior_phi)
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy,
        b=make_b_numpy,
        prior_mu=make_explicit_prior,
        n=make_explicit_n,
        seed=make_explicit_seed,
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results_with_explicit_prior_mu)
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy,
        b=make_b_numpy,
        prior_s=make_explicit_prior,
        n=make_explicit_n,
        seed=make_explicit_seed,
    )
    post = np.array(list(BayesTEstimationClass.get_posteriors().values()))
    assert np.allclose(post, make_get_posteriors_results_with_explicit_prior_s)


def test_BayesTEstimattion_outputs_correct_summaries(
    make_a_numpy,
    make_b_numpy,
    make_explicit_seed,
    make_quantile_summary,
    make_hdi_summary,
    make_get_rhat,
    make_infer_delta_probability_result,
    make_infer_delta_bayes_factor_result,
):
    # Initialisation
    BayesTEstimationClass = BayesTEstimation()
    BayesTEstimationClass.fit_posteriors(
        a=make_a_numpy, b=make_b_numpy, seed=make_explicit_seed
    )
    summary = np.array(
        BayesTEstimationClass.quantile_summary().iloc[:, [0, 1, 2]], dtype="float"
    )
    assert np.allclose(summary, make_quantile_summary)
    summary = np.array(
        BayesTEstimationClass.hdi_summary().iloc[:, [0, 1, 2]], dtype="float"
    )
    assert np.allclose(summary, make_hdi_summary)
    summary = np.array(BayesTEstimationClass.get_rhat()["rhat"])
    assert np.allclose(summary, make_get_rhat)
    p, i = BayesTEstimationClass.infer_delta_probability()
    assert np.isclose(p, make_infer_delta_probability_result[0])
    assert i == make_infer_delta_probability_result[1]
    bf, i = BayesTEstimationClass.infer_delta_bayes_factor()
    assert np.isclose(bf, make_infer_delta_bayes_factor_result[0])
    assert i == make_infer_delta_bayes_factor_result[1]


# Run bayestplotters tests


def test__get_centre_lines_runs_without_error(make_draw):
    try:
        _get_centre_lines(make_draw, method="hdi")
    except:
        raise pytest.fail()


def test__get_intervals_runs_with_hdi_without_error(make_draw):
    try:
        _get_intervals(make_draw, method="hdi", bounds=0.95)
    except:
        raise pytest.fail()


def test__get_intervals_runs_with_quantile_without_error(make_draw):
    try:
        _get_intervals(make_draw, method="quantile", bounds=[0.025, 0.975])
    except:
        raise pytest.fail()


def test__make_density_go_without_error(make_draw):
    try:
        _make_density_go(make_draw, name="dummy")
    except:
        raise pytest.fail()


def test__make_histogram_go_without_error(make_draw):
    try:
        _make_histogram_go(make_draw, name="dummy")
    except:
        raise pytest.fail()


def test__make_line_go_with_without_error(make_draw):
    try:
        cl = _get_centre_lines(make_draw, method="hdi")
        _make_line_go(cl, name="dummy")
    except:
        raise pytest.fail()


def test__make_area_go_without_error(make_draw):
    try:
        intervals = _get_intervals(make_draw, method="hdi", bounds=0.95)
        _make_area_go(intervals, name="dummy")
    except:
        raise pytest.fail()


# Run plot_posterior method test


def test_plot_posterior_without_error(make_a_numpy, make_b_numpy, make_explicit_seed):
    try:
        # Initialisation
        BayesTEstimationClass = BayesTEstimation()
        BayesTEstimationClass.fit_posteriors(
            a=make_a_numpy, b=make_b_numpy, seed=make_explicit_seed
        )
        BayesTEstimationClass.posterior_plot()
    except:
        raise pytest.fail()
